use std::{
  cell::{OnceCell, RefCell},
  pin::Pin,
  rc::Rc,
  sync::{mpsc, Arc, RwLock},
};

use disarm64::decoder;
use jni::{JNIEnv, JavaVM};
use num_derive::FromPrimitive;
use num_traits::FromPrimitive;
use unicorn_engine::{Arch, ArmCpRegister, HookType, Mode, Permission, Query, RegisterARM64, Unicorn};
use zerocopy::IntoBytes;

use crate::{component::{mmio::{calculate_components_start, component_mmio_read, component_mmio_write, ComponentMmioControl}, Component, ComponentIoResult}, computer::Computer, unsync_cell::UnsyncCell};

pub struct CoreHandle {
  message_sender: mpsc::Sender<CpuMessage>,
}

/*



todo REWRITEEEEEEE RAAAAH



 */

pub struct CpuContext<'a> {
  component_data_start: usize,
  components: Arc<[RwLock<Component>]>,
  /// see comment on [Computer]
  component_control: Arc<UnsyncCell<ComponentMmioControl>>,
  message_sender: mpsc::Sender<CpuMessage>,
  component_access_futures: Vec<Pin<Box<dyn Future<Output = ComponentIoResult>>>>,
  ram: Arc<UnsyncCell<[u8]>>,
  stopped: bool,
  _env: JNIEnv<'a>,
}

enum CpuMessage {
  Tick,
  Wake(Option<u64>),
  Sleep,
}

impl CoreHandle {
  pub fn new(computer: &Computer, core_index: usize, instructions_per_tick: u32, java_vm: JavaVM) -> Self {
    let (components, component_control) = (computer.components.clone(), computer.component_control.clone());
    let (started_up_sender, started_up_receiver) = oneshot::channel();

    let (message_sender, message_receiver) = mpsc::channel();
    let memories = (computer.shared_ram.clone(), computer.shared_rom.clone());

    std::thread::spawn({
      let message_sender = message_sender.clone();

      move || {
        let env = java_vm.attach_current_thread_permanently().unwrap();

        let context = CpuContext {
          component_data_start: calculate_components_start(components.len()),
          components,
          component_control,
          message_sender,
          component_access_futures: Vec::new(),
          ram: memories.0.clone(),
          _env: env,
          stopped: true,
        };
        thread(
          core_index,
          instructions_per_tick,
          memories,
          context,
          started_up_sender,
          message_receiver,
        )
      }
    });

    let _ = started_up_receiver.recv().expect("failed to start up core");

    CoreHandle { message_sender }
  }

  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_RustNative_coreFree")]
  unsafe extern "C" fn free<'local>(_env: JNIEnv<'local>, core: *const Self) {
    // Safety: the responsibility of verifying the pointer is valid is on the java caller
    drop(unsafe { Arc::from_raw(core) })
  }

  pub fn tick(&self) {
    self.message_sender.send(CpuMessage::Tick).expect("cpu thread panicked :(");
  }

  pub fn wake_at(&self, start: u64) {
    self.message_sender.send(CpuMessage::Wake(Some(start))).unwrap();
  }
}

#[derive(Debug, FromPrimitive)]
enum Exception {
  UndefinedInstruction = 1,
  SupervisorCall = 2,
  PrefetchAbort = 3,
  DataAbort = 4,
  Interrupt = 5,
  FastInterrupt = 6,
  Breakpoint = 7,
  ExceptionExit = 8,
  KernelTrap = 9,
  HypervisorCall = 11,
  HypervisorTrap = 12,
  SecureMonitorCall = 13,
  VirtualInterrupt = 14,
  VirtualFastInterrupt = 15,
  Semihost = 16,
  Nocp = 17,
  Invstate = 18,
  Stkof = 19,
  Lazyfp = 20,
  Lserr = 21,
  Unaligned = 22,
}

fn thread(
  core_index: usize,
  instructions_per_tick: u32,
  (shared_ram, shared_rom): (Arc<UnsyncCell<[u8]>>, Arc<UnsyncCell<[u8]>>),
  context: CpuContext,
  started_up_sender: oneshot::Sender<()>,
  message_receiver: mpsc::Receiver<CpuMessage>,
) -> Result<(), ()> {
  let mut vcpu =
    Unicorn::new_with_data(Arch::ARM64, Mode::LITTLE_ENDIAN, context).expect("failed to spawn unicorn engine");

  const MPIDR_EL1: ArmCpRegister = ArmCpRegister::from_manual(0b11, 0b000, 0b0000, 0b0000, 0b101);
  vcpu.reg_write_cp(CP_REG, MPIDR_EL1, core_index as u64).unwrap();

  vcpu
    .mmio_map(
      0x2000_0000,
      0x1000,
      Some(component_mmio_read),
      Some(component_mmio_write),
    )
    .expect("failed to map mmio for core");
  shared_rom
    .map(&mut vcpu, 0x4000_0000, None, Permission::READ | Permission::EXEC)
    .expect("failed to map memory for rom");
  shared_ram.map(&mut vcpu, 0x8000_0000, None, Permission::ALL).expect("failed to map memory for rom");
  use RegisterARM64::*;

  vcpu
    .add_intr_hook(|vcpu, excp| {
      println!("interrupt: {excp}");
      let exception = Exception::from_u32(excp).unwrap();

      let pc = vcpu.pc_read().unwrap();
      println!("exception: {exception:?}");
      let mut opcode = 0u32;
      if let Ok(_) = vcpu.mem_read(pc, opcode.as_mut_bytes()) {
        if let Some(opcode) = decoder::decode(opcode) {
          let mut inst = String::new();
          disarm64::format_insn::format_insn_pc(pc, &mut inst, &opcode).unwrap();
          println!("pc {pc:08X} {inst}");
        } else {
          println!("pc {pc:08X} - op parse failed {opcode:08X}");
        }
      } else {
        println!("pc {pc:08X} - can't read");
      }
      println!("x0 {:08X}", vcpu.reg_read(X0).unwrap());
      println!("x1 {:08X}", vcpu.reg_read(X1).unwrap());

      let syndrome = vcpu.query(Query::SYNDROME).unwrap();
      println!("0x{syndrome:08X}");
      vcpu.get_data_mut().stopped = true;
      vcpu.emu_stop().unwrap();
    })
    .unwrap();

  vcpu
    .add_mem_hook(
      HookType::MEM_INVALID,
      u64::MIN,
      u64::MAX,
      |vcpu, mem_type, address, _, value| {
        eprintln!("memory exception: {mem_type:?} at {address:X} (value: {value:X})");

        vcpu.emu_stop().unwrap();
        true
      },
    )
    .unwrap();

  started_up_sender.send(()).expect("panicked when trying to core");
  let try_fetch_message = |stopped| {
    if stopped {
      message_receiver.recv().map(Some).map_err(|_| ())
    } else {
      Ok(message_receiver.try_recv().ok())
    }
  };

  let mut instructions_left = 0;
  loop {
    while let Some(message) = try_fetch_message(vcpu.get_data().stopped || instructions_left == 0)? {
      match message {
        CpuMessage::Tick => {
          if vcpu.get_data().stopped {
            continue;
          }
        }
        CpuMessage::Wake(address) => {
          if let Some(address) = address {
            vcpu.set_pc(address).unwrap();
            println!("waking at {address}");
          }
          vcpu.get_data_mut().stopped = false;

          let pc = vcpu.pc_read().unwrap();
          println!("executing {pc:X}");
          vcpu.emu_start(pc, 0, 0, instructions_per_tick as usize).expect("error while executing cpu");
        }
        CpuMessage::Sleep => vcpu.get_data_mut().stopped = true,
      }
    }
    vcpu.emu_start(vcpu.pc_read().unwrap(), 0, 0, instructions_left.min(1000)).expect("error while executing cpu");
    instructions_left = instructions_left.saturating_sub(1000);
  }
}
